//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2024 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef f_AT_PRINTEROUTPUT_H
#define f_AT_PRINTEROUTPUT_H

#include <vd2/system/atomic.h>
#include <vd2/system/function.h>
#include <vd2/system/VDString.h>
#include <vd2/system/vectors.h>
#include <at/atcore/deviceprinter.h>
#include <at/atcore/notifylist.h>

class ATPrinterOutputManager;

// Printer output interface for text output. This gets presented in the
// printer view as a text edit widget.
//
class ATPrinterOutput final : public IATPrinterOutput {
	ATPrinterOutput(const ATPrinterOutput&) = delete;
	ATPrinterOutput& operator=(const ATPrinterOutput&) = delete;
public:
	ATPrinterOutput(ATPrinterOutputManager& parent);
	~ATPrinterOutput();

	void SetOnInvalidation(vdfunction<void()> fn);
	void Revalidate();

	size_t GetLength() const;
	const wchar_t *GetTextPointer(size_t offset) const;

	void Clear();

public:
	int AddRef() override;
	int Release() override;
	void *AsInterface(uint32 id) override;

	bool WantUnicode() const override;
	void WriteRaw(const uint8 *buf, size_t len) override;
	void WriteUnicode(const wchar_t *buf, size_t len) override;

private:
	static constexpr size_t kMaxTextLength = 0x7F000000;

	VDStringW mText;
	uint32 mColumn = 0;
	uint8 mDropNextChar = 0;
	bool mbInvalidated = false;

	vdfunction<void()> mpOnInvalidationFn;

	VDAtomicInt mRefCount{0};
	ATPrinterOutputManager& mParent;
};

// Printer output handler for graphical output, produced by dot matrix patterns.
// This gets presented in the printer view as a rasterized image.
//
class ATPrinterGraphicalOutput final : public IATPrinterGraphicalOutput {
	ATPrinterGraphicalOutput(const ATPrinterGraphicalOutput&) = delete;
	ATPrinterGraphicalOutput& operator=(const ATPrinterGraphicalOutput&) = delete;
public:
	ATPrinterGraphicalOutput(ATPrinterOutputManager& parent, const ATPrinterGraphicsSpec& spec);
	~ATPrinterGraphicalOutput();

	const ATPrinterGraphicsSpec& GetGraphicsSpec() const;
	vdrect32f GetDocumentBounds() const;

	void Clear();

	void SetOnInvalidation(vdfunction<void()> fn);
	bool ExtractInvalidationRect(bool& all, vdrect32f& r);

	struct CullInfo {
		size_t mLineStart;
		size_t mLineEnd;
	};

	bool PreCull(CullInfo& cullInfo, const vdrect32f& r) const;

	struct RenderDot {
		float mX;
		float mY;
	};

	void ExtractNextLineDots(vdfastvector<RenderDot>& renderDots, CullInfo& cullInfo, const vdrect32f& r) const;

public:
	int AddRef() override;
	int Release() override;
	void *AsInterface(uint32 id) override;

	void FeedPaper(float distanceMM) override;
	void Print(float x, uint32 dots) override;

private:
	struct Line {
		float mY = 0;
		uint32 mColumnStart = 0;
		uint32 mColumnCount = 0;
	};

	struct LineCompareY {
		bool operator()(const Line& line, float y) {
			return line.mY < y;
		}
	};

	struct PrintColumn {
		float mX = 0;
		uint32 mDots = 0;
	};

	float mPageWidthMM = 0;
	float mPageVBorderMM = 0;
	float mDotSizeMM = 0;
	float mHeadY = 0;
	float mHeadFirstBitOffsetY = 0;
	float mDotStepY = 0;
	float mHeadWidth = 0;
	float mHeadHeight = 0;
	int mHeadPinCount = 0;

	Line *mpCurrentLine = nullptr;
	vdfastvector<Line> mLines;
	vdfastvector<PrintColumn> mColumns;

	bool mbInvalidated = false;
	bool mbInvalidatedAll = false;
	vdrect32f mInvalidationRect;
	vdfunction<void()> mpOnInvalidationFn;

	const ATPrinterGraphicsSpec mGraphicsSpec;

	VDAtomicInt mRefCount{0};
	ATPrinterOutputManager& mParent;
};

class ATPrinterOutputManager final : public vdrefcounted<IATPrinterOutputManager> {
	ATPrinterOutputManager(const ATPrinterOutputManager&);
	ATPrinterOutputManager& operator=(const ATPrinterOutputManager&);
public:
	ATPrinterOutputManager();
	~ATPrinterOutputManager();

	uint32 GetOutputCount() const;
	uint32 GetGraphicalOutputCount() const;

	ATPrinterOutput& GetOutput(uint32 idx) const;
	ATPrinterGraphicalOutput& GetGraphicalOutput(uint32 idx) const;

	ATNotifyList<const vdfunction<void(ATPrinterOutput&)> *> OnAddedOutput;
	ATNotifyList<const vdfunction<void(ATPrinterOutput&)> *> OnRemovingOutput;
	ATNotifyList<const vdfunction<void(ATPrinterGraphicalOutput&)> *> OnAddedGraphicalOutput;
	ATNotifyList<const vdfunction<void(ATPrinterGraphicalOutput&)> *> OnRemovingGraphicalOutput;

public:
	vdrefptr<IATPrinterOutput> CreatePrinterOutput() override;
	vdrefptr<IATPrinterGraphicalOutput> CreatePrinterGraphicalOutput(const ATPrinterGraphicsSpec& spec) override;

public:
	void OnDestroyingOutput(ATPrinterOutput& output);
	void OnDestroyingOutput(ATPrinterGraphicalOutput& output);

private:
	vdfastvector<ATPrinterOutput *> mOutputs;
	vdfastvector<ATPrinterGraphicalOutput *> mGraphicalOutputs;
};

#endif
