//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2024 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <vd2/system/bitmath.h>
#include <vd2/system/Error.h>
#include <vd2/system/vdstl_vectorview.h>
#include "printeroutput.h"

ATPrinterOutput::ATPrinterOutput(ATPrinterOutputManager& parent)
	: mParent(parent)
{
}

ATPrinterOutput::~ATPrinterOutput() {
}

void ATPrinterOutput::SetOnInvalidation(vdfunction<void()> fn) {
	mpOnInvalidationFn = std::move(fn);
}

void ATPrinterOutput::Revalidate() {
	mbInvalidated = false;
}

size_t ATPrinterOutput::GetLength() const {
	return mText.size();
}

const wchar_t *ATPrinterOutput::GetTextPointer(size_t offset) const {
	return mText.c_str() + offset;
}

void ATPrinterOutput::Clear() {
	mText.clear();
	mColumn = 0;
}

int ATPrinterOutput::AddRef() {
	return ++mRefCount;
}

int ATPrinterOutput::Release() {
	const int rc = --mRefCount;
	if (!rc) {
		mParent.OnDestroyingOutput(*this);
		delete this;
	}

	return rc;
}

void *ATPrinterOutput::AsInterface(uint32 id) {
	if (id == IATPrinterOutput::kTypeID)
		return static_cast<IATPrinterOutput *>(this);

	return nullptr;
}

bool ATPrinterOutput::WantUnicode() const {
	return true;
}

void ATPrinterOutput::WriteRaw(const uint8 *buf, size_t len) {
	if (!len)
		return;

	const size_t startPos = mText.size();

	if (startPos >= kMaxTextLength)
		return;

	while(len--) {
		uint8 c = *buf++;

		if (c != 0x0A && c != 0x0D) {
			c &= 0x7f;
			
			if (c < 0x20 || c > 0x7F)
				c = '?';
		}

		if (mDropNextChar) {
			if (mDropNextChar == c) {
				mDropNextChar = 0;

				continue;
			}

			mDropNextChar = 0;
		}

		if (c == 0x0D || c == 0x0A) {
			mDropNextChar = c ^ (uint8)(0x0D ^ 0x0A);
			mText.push_back(0x0A);
			mColumn = 0;
		} else {
			mText.push_back(c);
			++mColumn;

			if (mColumn >= 132) {
				c = 0x0A;
				mText.push_back(c);
				mColumn = 0;
			}
		}
	}

	if (!mbInvalidated && mText.size() != startPos) {
		mbInvalidated = true;

		if (mpOnInvalidationFn)
			mpOnInvalidationFn();
	}
}

void ATPrinterOutput::WriteUnicode(const wchar_t *buf, size_t len) {
	if (!len)
		return;

	const size_t startPos = mText.size();

	while(len--) {
		wchar_t ch = *buf++;

		if (!ch)
			continue;

		mText.push_back(ch);

		if (ch == '\n')
			mColumn = 0;
		else if (++mColumn >= 132) {
			mColumn = 0;
			ch = '\n';
			mText.push_back(ch);
		}
	}

	if (!mbInvalidated && mText.size() != startPos) {
		mbInvalidated = true;

		if (mpOnInvalidationFn)
			mpOnInvalidationFn();
	}
}

////////////////////////////////////////////////////////////////////////////////

ATPrinterGraphicalOutput::ATPrinterGraphicalOutput(ATPrinterOutputManager& parent, const ATPrinterGraphicsSpec& spec)
	: mParent(parent)
	, mGraphicsSpec(spec)
{
	mPageWidthMM = spec.mPageWidthMM;
	mPageVBorderMM = spec.mPageVBorderMM;
	mDotSizeMM = spec.mDotSizeMM;

	mHeadFirstBitOffsetY = spec.mDotSizeMM;
	if (!spec.mbBit0Top)
		mHeadFirstBitOffsetY += spec.mVerticalDotPitchMM * (spec.mNumPins - 1);

	mDotStepY = spec.mbBit0Top ? spec.mVerticalDotPitchMM : -spec.mVerticalDotPitchMM;
	mHeadWidth = spec.mDotSizeMM * 2;
	mHeadHeight = spec.mDotSizeMM * 2 + spec.mVerticalDotPitchMM * (spec.mNumPins - 1);
	mHeadPinCount = spec.mNumPins;

	Clear();
}

ATPrinterGraphicalOutput::~ATPrinterGraphicalOutput() {
}

const ATPrinterGraphicsSpec& ATPrinterGraphicalOutput::GetGraphicsSpec() const {
	return mGraphicsSpec;
}

vdrect32f ATPrinterGraphicalOutput::GetDocumentBounds() const {
	vdrect32f bounds(0, 0, 0, 0);

	if (!mLines.empty()) {
		bounds.left = 0;
		bounds.right = mPageWidthMM;
		bounds.top = 0;
		bounds.bottom = mLines.back().mY + mHeadHeight + mPageVBorderMM;
	}

	return bounds;
}

void ATPrinterGraphicalOutput::Clear() {
	mHeadY = mPageVBorderMM + mHeadFirstBitOffsetY;

	mpCurrentLine = nullptr;
	mLines.clear();
	mColumns.clear();

	mbInvalidatedAll = true;

	if (!mbInvalidated) {
		mbInvalidated = true;

		if (mpOnInvalidationFn)
			mpOnInvalidationFn();
	}
}

void ATPrinterGraphicalOutput::SetOnInvalidation(vdfunction<void()> fn) {
	mpOnInvalidationFn = std::move(fn);
}

bool ATPrinterGraphicalOutput::ExtractInvalidationRect(bool& all, vdrect32f& r) {
	all = false;
	r = {};

	if (!mbInvalidated)
		return false;

	mbInvalidated = false;

	if (mbInvalidatedAll) {
		mbInvalidatedAll = false;
		all = true;
	} else {
		r = mInvalidationRect;
	}

	return true;
}

bool ATPrinterGraphicalOutput::PreCull(CullInfo& cullInfo, const vdrect32f& r) const {
	auto itBegin = mLines.begin();
	auto itLine1 = std::lower_bound(itBegin, mLines.end(), r.top - mHeadHeight, LineCompareY());
	auto itLine2 = std::lower_bound(itLine1, mLines.end(), r.bottom, LineCompareY());

	cullInfo.mLineStart = (size_t)(itLine1 - itBegin);
	cullInfo.mLineEnd = (size_t)(itLine2 - itBegin);

	return cullInfo.mLineEnd > cullInfo.mLineStart;
}

void ATPrinterGraphicalOutput::ExtractNextLineDots(vdfastvector<RenderDot>& renderDots, CullInfo& cullInfo, const vdrect32f& r) const {
	const auto itLine1 = mLines.begin() + cullInfo.mLineStart;
	const auto itLine2 = mLines.begin() + cullInfo.mLineEnd;

	const auto itRowLine1 = std::lower_bound(itLine1, itLine2, r.top - mHeadHeight, LineCompareY());
	const auto itRowLine2 = std::lower_bound(itRowLine1, itLine2, r.bottom, LineCompareY());

	const float dotRadius = mDotSizeMM;
	const float docRowY1 = r.top;
	const float docRowY2 = r.bottom;
	const float viewDocXC = (r.right + r.left) * 0.5f;
	const float viewDocXD = (r.right - r.left) * 0.5f;

	for(auto itLine = itRowLine1; itLine != itRowLine2; ++itLine) {
		const uint32 columnIdx = itLine->mColumnStart;
		const float firstDotY = itLine->mY + mHeadFirstBitOffsetY;
			
		// pre-coll pins on the print head -- the tricky part is that the head may be upside-down
		uint32 dotMask = (UINT32_C(2) << (this->mHeadPinCount - 1)) - 1;

		if (mDotStepY > 0) {
			// check if first dot is entirely above the clip rect
			const float firstDotY2 = firstDotY + dotRadius;
			if (firstDotY2 <= docRowY1) {
				// compute how many dots to trim off
				int topDotsToCull = 1 + (int)((docRowY1 - firstDotY2) / mDotStepY);

				if (topDotsToCull >= mHeadPinCount)
					dotMask = 0;
				else
					dotMask &= dotMask << topDotsToCull;
			}

			// check if the last dot is entirely below the clip rect
			const float lastDotY1 = firstDotY + mDotStepY * (float)(mHeadPinCount - 1) - dotRadius;
			if (lastDotY1 >= docRowY2) {
				// compute how many dots to trim off
				int bottomDotsToCull = 1 + (int)((lastDotY1 - docRowY2) / mDotStepY);

				if (bottomDotsToCull >= mHeadPinCount)
					dotMask = 0;
				else
					dotMask &= dotMask >> bottomDotsToCull;
			}
		} else if (mDotStepY < 0) {
			// check if first dot is entirely below the clip rect
			const float firstDotY1 = firstDotY - dotRadius;
			if (firstDotY1 >= docRowY2) {
				// compute how many dots to trim off
				int bottomDotsToCull = 1 + (int)((firstDotY1 - docRowY2) / mDotStepY);

				if (bottomDotsToCull >= mHeadPinCount)
					dotMask = 0;
				else
					dotMask >>= bottomDotsToCull;
			}

			// check if the last dot is entirely above the clip rect
			const float lastDotY2 = firstDotY + mDotStepY * (float)(mHeadPinCount - 1) + dotRadius;
			if (lastDotY2 <= docRowY1) {
				// compute how many dots to trim off
				int topDotsToCull = 1 + (int)((docRowY1 - lastDotY2) / mDotStepY);

				if (topDotsToCull >= mHeadPinCount)
					dotMask = 0;
				else
					dotMask &= dotMask << topDotsToCull;
			}
		}

		if (!dotMask)
			continue;
			
		for(const PrintColumn& column : vdvector_view(mColumns.data() + columnIdx, itLine->mColumnCount)) {
			if (fabsf(column.mX - viewDocXC) >= viewDocXD + dotRadius)
				continue;

			uint32 pins = column.mDots & dotMask;

			while(pins) {
				int idx = VDFindLowestSetBitFast(pins);
				pins &= (pins - 1);

				const float dotY = firstDotY + mDotStepY * (float)idx;

				renderDots.push_back(
					RenderDot {
						.mX = column.mX,
						.mY = dotY,
					}
				);
			}
		}
	}
}

int ATPrinterGraphicalOutput::AddRef() {
	return ++mRefCount;
}

int ATPrinterGraphicalOutput::Release() {
	const int rc = --mRefCount;
	if (!rc) {
		mParent.OnDestroyingOutput(*this);
		delete this;
	}

	return rc;
}

void *ATPrinterGraphicalOutput::AsInterface(uint32 id) {
	if (id == IATPrinterGraphicalOutput::kTypeID)
		return static_cast<IATPrinterGraphicalOutput *>(this);

	return nullptr;
}

void ATPrinterGraphicalOutput::FeedPaper(float distanceMM) {
	mHeadY += distanceMM;
	mpCurrentLine = nullptr;
}

void ATPrinterGraphicalOutput::Print(float x, uint32 pins) {
	// if no dots, ignore
	if (!pins)
		return;

	// if no line, establish now
	if (!mpCurrentLine) {
		Line newLine {};
		newLine.mY = mHeadY - mDotSizeMM;
		newLine.mColumnStart = (uint32)mColumns.size();
		newLine.mColumnCount = 0;

		auto itLine = std::lower_bound(
			mLines.begin(),
			mLines.end(), 
			newLine,
			[](const Line& a, const Line& b) {
				return a.mY < b.mY;
			}
		);

		mpCurrentLine = &*mLines.insert(itLine, newLine);
	}

	auto& column = mColumns.emplace_back();
	column.mX = x;
	column.mDots = pins;

	++mpCurrentLine->mColumnCount;

	if (!mbInvalidatedAll) {
		vdrect32f r;
		r.left = x - mDotSizeMM;
		r.top = mHeadY - mDotSizeMM;
		r.right = r.left + mHeadWidth;
		r.bottom = r.top + mHeadHeight;

		if (!mbInvalidated) {
			mbInvalidated = true;

			mInvalidationRect = r;

			if (mpOnInvalidationFn)
				mpOnInvalidationFn();
		} else {
			mInvalidationRect.left = std::min(mInvalidationRect.left, r.left);
			mInvalidationRect.top = std::min(mInvalidationRect.top, r.top);
			mInvalidationRect.right = std::max(mInvalidationRect.right, r.right);
			mInvalidationRect.bottom = std::max(mInvalidationRect.bottom, r.bottom);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

ATPrinterOutputManager::ATPrinterOutputManager() {
}

ATPrinterOutputManager::~ATPrinterOutputManager() {
}

uint32 ATPrinterOutputManager::GetOutputCount() const {
	return (uint32)mOutputs.size();
}

uint32 ATPrinterOutputManager::GetGraphicalOutputCount() const {
	return (uint32)mGraphicalOutputs.size();
}

ATPrinterOutput& ATPrinterOutputManager::GetOutput(uint32 idx) const {
	return *mOutputs[idx];
}

ATPrinterGraphicalOutput& ATPrinterOutputManager::GetGraphicalOutput(uint32 idx) const {
	return *mGraphicalOutputs[idx];
}

vdrefptr<IATPrinterOutput> ATPrinterOutputManager::CreatePrinterOutput() {
	vdrefptr<ATPrinterOutput> output(new ATPrinterOutput(*this));

	mOutputs.push_back(output);
	OnAddedOutput.InvokeAll(*output);

	return output;
}

vdrefptr<IATPrinterGraphicalOutput> ATPrinterOutputManager::CreatePrinterGraphicalOutput(const ATPrinterGraphicsSpec& spec) {
	vdrefptr<ATPrinterGraphicalOutput> output(new ATPrinterGraphicalOutput(*this, spec));

	mGraphicalOutputs.push_back(output);
	OnAddedGraphicalOutput.InvokeAll(*output);

	return output;
}

void ATPrinterOutputManager::OnDestroyingOutput(ATPrinterOutput& output) {
	auto it = std::find(mOutputs.begin(), mOutputs.end(), &output);

	if (it == mOutputs.end())
		VDRaiseInternalFailure();

	mOutputs.erase(it);

	OnRemovingOutput.InvokeAll(output);
}

void ATPrinterOutputManager::OnDestroyingOutput(ATPrinterGraphicalOutput& output) {
	auto it = std::find(mGraphicalOutputs.begin(), mGraphicalOutputs.end(), &output);

	if (it == mGraphicalOutputs.end())
		VDRaiseInternalFailure();

	mGraphicalOutputs.erase(it);

	OnRemovingGraphicalOutput.InvokeAll(output);
}

////////////////////////////////////////////////////////////////////////////////

vdrefptr<IATPrinterOutputManager> ATCreatePrinterOutputManager() {
	return vdrefptr<IATPrinterOutputManager>(new ATPrinterOutputManager);
}
