//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2022 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include "stdafx.h"
#include <windows.h>
#include <windowsx.h>
#include <vd2/system/binary.h>
#include <vd2/system/bitmath.h>
#include <vd2/system/color.h>
#include <vd2/system/cpuaccel.h>
#include <vd2/system/vdstl_vectorview.h>
#include <vd2/Dita/services.h>
#include <at/atcore/atascii.h>
#include <at/atnativeui/canvas_win32.h>
#include <at/atnativeui/uinativewindow.h>
#include "oshelper.h"
#include "printeroutput.h"
#include "resource.h"
#include "simulator.h"
#include "texteditor.h"
#include "uidbgprinteroutput.h"

extern ATSimulator g_sim;

///////////////////////////////////////////////////////////////////////////////

class ATUIPrinterGraphicalOutputWindow final : public ATUINativeWindow {
public:
	ATUIPrinterGraphicalOutputWindow();

	void AttachToOutput(ATPrinterGraphicalOutput& output);

	void Clear();

	void ResetView();
	void SaveAsPNG(float dpi);

private:
	LRESULT WndProc(UINT msg, WPARAM wParam, LPARAM lParam) override;
	LRESULT OnNCCalcSize(WPARAM wParam, LPARAM lParam);
	void OnDestroy();
	void OnSize();
	void OnKeyDown(uint32 vk, uint32 flags);
	void OnMouseMove(int x, int y);
	void OnMouseDownL(int x, int y);
	void OnMouseUpL(int x, int y);
	void OnMouseWheel(int x, int y, float delta);
	void OnCaptureChanged();
	bool OnGesture(WPARAM wParam, LPARAM lParam);
	void OnPaint();
	void OnDpiChanged(int dpi);

	void UpdateViewOrigin();
	void ScrollByPixels(int dx, int dy);
	void SetZoom(float clicks);
	void SetZoom(float clicks, const vdpoint32& centerPt);
	void UpdateZoom(const vdpoint32 *centerPt);

	void ForceFullInvalidation();
	void ProcessPendingInvalidation();
	void InvalidatePaperArea(float x1, float y1, float x2, float y2);

	vdrect32f GetDocumentBounds() const;

	struct ViewTransform {
		float mOriginX = 0;
		float mOriginY = 0;
		float mPixelsPerMM = 1;
		float mMMPerPixel = 1;
	};

	bool Render(const ViewTransform& viewTransform, sint32 x, sint32 y, uint32 w, uint32 h, bool force);
	void Downsample8x8(uint32 *dst, const uint8 *src, size_t w);
	void Downsample8x8_Scalar(uint32 *dst, const uint8 *src, size_t w);
#if defined(VD_CPU_X64)
	void Downsample8x8_POPCNT64(uint32 *dst, const uint8 *src, size_t w);
#elif defined(VD_CPU_X86)
	void Downsample8x8_POPCNT32(uint32 *dst, const uint8 *src, size_t w);
#endif

	float mPageWidthMM = 0;
	float mPageVBorderMM = 0;
	float mDotSizeMM = 0;

	float mViewOriginX = 0;
	float mViewOriginY = 0;
	float mViewCenterX = 0;
	float mViewCenterY = 0;
	float mViewPixelsPerMM = 1.0f;
	float mViewMMPerPixel = 1.0f;
	sint32 mViewWidthPixels = 0;
	sint32 mViewHeightPixels = 0;
	int mViewDpi = 96;

	bool mbDragging = false;
	sint32 mDragLastX = 0;
	sint32 mDragLastY = 0;
	float mWheelAccum = 0;

	bool mbInGesture = false;
	bool mbFirstGestureEvent = false;
	vdpoint32 mGestureOrigin;
	float mGestureZoomOrigin = 0;

	static constexpr float kZoomMin = -10.0f;
	static constexpr float kZoomMax = 25.0f;
	float mZoomClicks = 0;

	bool mbInvalidationPending = false;

	using RenderDot = ATPrinterGraphicalOutput::RenderDot;
	vdfastvector<RenderDot> mDotCullBuffer;

	ATPrinterGraphicalOutput *mpOutput = nullptr;
	ATUICanvasW32 mCanvas;
	vdfastvector<uint32> mFrameBuffer;
	vdfastvector<uint8> mABuffer;

	vdstructex<RGNDATA> mRegionData;

	uint8 mPopCnt8[256];
	uint32 mGammaTable[65];
};

ATUIPrinterGraphicalOutputWindow::ATUIPrinterGraphicalOutputWindow() {
	for(int i=0; i<256; ++i) {
		uint8 v = i;

		v -= (v >> 1) & 0x55;
		v = (v & 0x33) + ((v & 0xCC) >> 2);

		mPopCnt8[i] = (v & 0x0F) + (v >> 4);
	}

	for(int i=0; i<=64; ++i) {
		float luma = 0.90f - 0.87f * (float)i / 64.0f;
		uint32 v = VDColorRGB(vdfloat32x3::set(luma, luma, luma)).LinearToSRGB().ToRGB8();

		mGammaTable[i] = v;
	}
}

void ATUIPrinterGraphicalOutputWindow::AttachToOutput(ATPrinterGraphicalOutput& output) {
	const ATPrinterGraphicsSpec& spec = output.GetGraphicsSpec();

	mpOutput = &output;
	mpOutput->SetOnInvalidation(
		[this] {
			if (!mbInvalidationPending) {
				mbInvalidationPending = true;

				::PostMessage(mhwnd, WM_USER + 100, 0, 0);
			}
		}
	);

	mPageWidthMM = spec.mPageWidthMM;
	mPageVBorderMM = spec.mPageVBorderMM;
	mDotSizeMM = spec.mDotSizeMM;

	ResetView();
}

void ATUIPrinterGraphicalOutputWindow::Clear() {
	mViewCenterY = 0;

	UpdateViewOrigin();

	if (mpOutput)
		mpOutput->Clear();
}

void ATUIPrinterGraphicalOutputWindow::ResetView() {
	mZoomClicks = -1000.0f;
	SetZoom(0);

	mViewCenterX = mPageWidthMM * 0.5f;
	mViewCenterY = mPageVBorderMM;

	UpdateViewOrigin();

	ForceFullInvalidation();
}

void ATUIPrinterGraphicalOutputWindow::SaveAsPNG(float dpi) {
	const VDStringW& fn = VDGetSaveFileName("PrinterSaveAsPNG"_vdtypeid, (VDGUIHandle)mhwnd, L"Save PNG image", L"PNG image\0*.png\0", L"png");
	if (fn.empty())
		return;

	// get bounds in document space
	vdrect32f documentBounds = GetDocumentBounds();

	// compute pixel bounds
	const float mmToInches = 1.0f / 25.4f;
	const sint32 w = std::max<sint32>(1, (sint32)ceilf(documentBounds.width() * mmToInches * dpi));
	const sint32 h = std::max<sint32>(1, (sint32)ceilf(documentBounds.height() * mmToInches * dpi));

	// render whole screen
	ViewTransform vt;
	vt.mOriginX = documentBounds.left;
	vt.mOriginY = documentBounds.top;
	vt.mPixelsPerMM = mmToInches * dpi;
	vt.mMMPerPixel = 1.0f / vt.mPixelsPerMM;
	Render(vt, 0, 0, w, h, true);

	// save as PNG
	VDPixmap px {};
	px.w = w;
	px.h = h;
	px.data = &mFrameBuffer[w * (h - 1)];
	px.pitch = -w*4;
	px.format = nsVDPixmap::kPixFormat_XRGB8888;

	ATSaveFrame(px, fn.c_str());

	// clear out large buffers
	vdfastvector<uint32> fb;
	fb.swap(mFrameBuffer);

	vdfastvector<uint8> ab;
	ab.swap(mABuffer);
}

LRESULT ATUIPrinterGraphicalOutputWindow::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_CREATE:
			mCanvas.Init(mhwnd);
			SetTouchMode(kATUITouchMode_2DPanSmooth);
			break;

		case WM_DESTROY:
			OnDestroy();
			break;

		case WM_SIZE:
			OnSize();
			return 0;

		case WM_NCCALCSIZE:
			return OnNCCalcSize(wParam, lParam);

		case WM_KEYDOWN:
			OnKeyDown(wParam, lParam);
			return 0;

		case WM_MOUSEMOVE:
			OnMouseMove(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			Focus();
			OnMouseDownL(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_LBUTTONUP:
			OnMouseUpL(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_MOUSEWHEEL:
			OnMouseWheel(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), (float)GET_WHEEL_DELTA_WPARAM(wParam) / (float)WHEEL_DELTA);
			return 0;

		case WM_CAPTURECHANGED:
			OnCaptureChanged();
			return 0;

		case WM_PAINT:
			OnPaint();
			return 0;

		case WM_ERASEBKGND:
			return 0;

		case WM_GESTURE:
			if (OnGesture(wParam, lParam))
				return 0;
			break;

		case WM_DPICHANGED:
		case ATWM_INHERIT_DPICHANGED:
			OnDpiChanged(wParam);
			break;

		case WM_USER + 100:
			ProcessPendingInvalidation();
			return 0;
	}

	return ATUINativeWindow::WndProc(msg, wParam, lParam);
}

LRESULT ATUIPrinterGraphicalOutputWindow::OnNCCalcSize(WPARAM wParam, LPARAM lParam) {
	if (!wParam)
		return DefWindowProc(mhwnd, WM_NCCALCSIZE, wParam, lParam);

	NCCALCSIZE_PARAMS& params = *(NCCALCSIZE_PARAMS *)lParam;

	// ask DWP to compute the new client size
	DefWindowProc(mhwnd, WM_NCCALCSIZE, FALSE, (LPARAM)&params.rgrc[0]);

	// compute the source and destination client rect sizes
	sint32 neww = params.rgrc[0].right  - params.rgrc[0].left;
	sint32 newh = params.rgrc[0].bottom - params.rgrc[0].top;
	sint32 oldw = params.rgrc[2].right  - params.rgrc[2].left;
	sint32 oldh = params.rgrc[2].bottom - params.rgrc[2].top;

	// compute the difference in center offsets
	sint32 newcx = neww / 2;
	sint32 newcy = newh / 2;
	sint32 oldcx = oldw / 2;
	sint32 oldcy = oldh / 2;

	// compute the size to copy
	sint32 savecx = std::min<sint32>(oldcx, newcx);
	sint32 savecy = std::min<sint32>(oldcy, newcy);
	sint32 savew = std::min<sint32>(oldw, neww);
	sint32 saveh = std::min<sint32>(oldh, newh);

	// set source and dest copy rects
	params.rgrc[1].left   = params.rgrc[0].left + newcx - savecx;
	params.rgrc[1].top    = params.rgrc[0].top  + newcy - savecy;
	params.rgrc[1].right  = params.rgrc[1].left + savew;
	params.rgrc[1].bottom = params.rgrc[1].top  + saveh;

	params.rgrc[2].left   = params.rgrc[2].left + oldcx - savecx;
	params.rgrc[2].top    = params.rgrc[2].top  + oldcy - savecy;
	params.rgrc[2].right  = params.rgrc[2].left + savew;
	params.rgrc[2].bottom = params.rgrc[2].top  + saveh;

	return WVR_VALIDRECTS;
}

void ATUIPrinterGraphicalOutputWindow::OnDestroy() {
	if (mpOutput) {
		mpOutput->SetOnInvalidation(nullptr);
		mpOutput = nullptr;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnSize() {
	auto [w, h] = GetClientSize();

	if (mViewWidthPixels != w || mViewHeightPixels != h) {
		mViewWidthPixels = w;
		mViewHeightPixels = h;

		// Update the view origin based on the center and new size. We don't redraw
		// here as that's already done by the NCCALCSIZE correction.
		UpdateViewOrigin();
	}
}

void ATUIPrinterGraphicalOutputWindow::OnKeyDown(uint32 vk, uint32 flags) {
	const bool ctrl = GetKeyState(VK_CONTROL) < 0;

	switch(vk) {
		case VK_ESCAPE:
			if (ATGetUIPane(kATUIPaneId_Display))
				ATActivateUIPane(kATUIPaneId_Display, true);
			break;

		case VK_LEFT:
			ScrollByPixels(ctrl ? 1 : 100, 0);
			break;

		case VK_RIGHT:
			ScrollByPixels(ctrl ? -1 : -100, 0);
			break;

		case VK_UP:
			ScrollByPixels(0, ctrl ? 1 : 100);
			break;

		case VK_DOWN:
			ScrollByPixels(0, ctrl ? -1 : -100);
			break;

		case VK_PRIOR:
			ScrollByPixels(0, mViewHeightPixels);
			break;

		case VK_NEXT:
			ScrollByPixels(0, -mViewHeightPixels);
			break;

		case VK_OEM_PLUS:
			if (ctrl)
				SetZoom(mZoomClicks + 1.0f);
			break;

		case VK_OEM_MINUS:
			if (ctrl)
				SetZoom(mZoomClicks - 1.0f);
			break;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseMove(int x, int y) {
	if (mbDragging) {
		int dx = x - mDragLastX;
		int dy = y - mDragLastY;
		mDragLastX = x;
		mDragLastY = y;

		ScrollByPixels(dx, dy);
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseDownL(int x, int y) {
	mbDragging = true;
	mDragLastX = x;
	mDragLastY = y;
	::SetCapture(mhwnd);
}

void ATUIPrinterGraphicalOutputWindow::OnMouseUpL(int x, int y) {
	if (mbDragging) {
		mbDragging = false;
		::ReleaseCapture();
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseWheel(int x, int y, float delta) {
	mWheelAccum += delta;

	const int clicks = (int)floorf(mWheelAccum + 0.5f);

	if (clicks) {
		const vdpoint32& cpt = TransformScreenToClient(vdpoint32(x, y));
		SetZoom(mZoomClicks + (float)clicks, cpt);

		mWheelAccum -= (int)clicks;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnCaptureChanged() {
	mbDragging = false;
}

bool ATUIPrinterGraphicalOutputWindow::OnGesture(WPARAM wParam, LPARAM lParam) {
	GESTUREINFO gestureInfo {sizeof(GESTUREINFO)};

	const BOOL haveInfo = GetGestureInfo((HGESTUREINFO)lParam, &gestureInfo);
	if (!haveInfo)
		return false;

	CloseGestureInfoHandle((HGESTUREINFO)lParam);

	switch(gestureInfo.dwID) {
		case GID_BEGIN:
			mbInGesture = true;
			mbFirstGestureEvent = true;
			break;

		case GID_END:
			mbInGesture = false;
			mbFirstGestureEvent = false;
			break;

		default:
			if (mbInGesture) {
				vdpoint32 pt(gestureInfo.ptsLocation.x, gestureInfo.ptsLocation.y);

				switch(gestureInfo.dwID) {
					case GID_PAN:
						if (!mbFirstGestureEvent)
							ScrollByPixels(pt.x - mGestureOrigin.x, pt.y - mGestureOrigin.y);

						mGestureOrigin = pt;
						break;

					case GID_ZOOM:
						if (float distance = (float)gestureInfo.ullArguments; mbFirstGestureEvent) {
							mGestureZoomOrigin = distance;
						} else {
							if (distance > mGestureZoomOrigin * 1.10f) {
								mGestureZoomOrigin *= 1.1487f;
								SetZoom(mZoomClicks + 1.0f);
							} else if (distance < mGestureZoomOrigin / 1.10f) {
								mGestureZoomOrigin /= 1.1487f;
								SetZoom(mZoomClicks - 1.0f);
							}
						}
						break;
				}

				mbFirstGestureEvent = false;
			}
			break;
	}

	return false;
}

void ATUIPrinterGraphicalOutputWindow::OnPaint() {
	PAINTSTRUCT ps;
	if (HDC hdc = mCanvas.BeginDirect(ps, false)) {
		const sint32 w = ps.rcPaint.right - ps.rcPaint.left;
		const sint32 h = ps.rcPaint.bottom - ps.rcPaint.top;

		if (w > 0 && h > 0) {
			RECT *paintRects = &ps.rcPaint;
			size_t numPaintRects = 1;

			HRGN hrgnUpdate = ::CreateRectRgn(0, 0, 0, 0);
			if (hrgnUpdate) {
				if (::GetRandomRgn(hdc, hrgnUpdate, SYSRGN) > 0) {
					DWORD numBytes = ::GetRegionData(hrgnUpdate, 0, nullptr);

					if (numBytes) {
						if (mRegionData.size() < numBytes)
							mRegionData.resize(numBytes);

						if (::GetRegionData(hrgnUpdate, mRegionData.size(), mRegionData.data())) {
							paintRects = (RECT *)mRegionData->Buffer;
							numPaintRects = mRegionData->rdh.nCount;

							// convert all rects from screen to client coords
							for(size_t i = 0; i < numPaintRects; ++i) {
								RECT& r = paintRects[i];

								const auto& cr = TransformScreenToClient(vdrect32(r.left, r.top, r.right, r.bottom));

								r.left = cr.left;
								r.top = cr.top;
								r.right = cr.right;
								r.bottom = cr.bottom;
							}
						}
					}
				}
				::DeleteObject(hrgnUpdate);
			}

			while(numPaintRects--) {
				const RECT& paintRect = *paintRects++;

				// convert left and right to document coordinates
				const float viewfx1 = mViewOriginX + (float)paintRect.left * mViewMMPerPixel;
				const float viewfx2 = mViewOriginX + (float)paintRect.right * mViewMMPerPixel;

				// if we are full left/right, just clear to background color
				SetBkColor(hdc, 0x808080);
				SetBkMode(hdc, OPAQUE);

				if (viewfx2 <= 0 || viewfx1 >= mPageWidthMM) {
					ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &paintRect, L"", 0, nullptr);
				} else {
					// compute clipped page width within paint bounds, now that we know we won't
					// massively overflow
					sint32 paintx1 = paintRect.left;
					sint32 paintx2 = paintRect.right;
					sint32 painty1 = paintRect.top;
					sint32 painty2 = paintRect.bottom;

					if (viewfx1 < 0) {
						paintx1 = (sint32)ceilf(-mViewOriginX * mViewPixelsPerMM - 0.5f / 8.0f);

						RECT rLeft  = paintRect;
						rLeft.right = paintx1;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rLeft, L"", 0, nullptr);
					}

					if (viewfx2 > mPageWidthMM) {
						paintx2 = (sint32)ceilf((mPageWidthMM - mViewOriginX) * mViewPixelsPerMM - 7.5f / 8.0f);

						RECT rRight  = paintRect;
						rRight.left = paintx2;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rRight, L"", 0, nullptr);
					}

					// check for clipping at top of page
					const float viewfy1 = mViewOriginY + (float)paintRect.top * mViewMMPerPixel;
					if (viewfy1 < 0) {
						painty1 = (sint32)ceilf(-mViewOriginY * mViewPixelsPerMM - 0.5f / 8.0f);

						RECT rTop = paintRect;
						rTop.left = paintx1;
						rTop.right = paintx2;
						rTop.bottom = painty1;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rTop, L"", 0, nullptr);
					}

					const sint32 paintw = paintx2 - paintx1;
					const sint32 painth = painty2 - painty1;

					if (paintw > 0 && painth > 0) {
						ViewTransform viewTransform;
						viewTransform.mOriginX = mViewOriginX;
						viewTransform.mOriginY = mViewOriginY;
						viewTransform.mMMPerPixel = mViewMMPerPixel;
						viewTransform.mPixelsPerMM = mViewPixelsPerMM;

						if (Render(viewTransform, paintx1, painty1, paintw, painth, false)) {
							BITMAPINFO bi {};
							bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
							bi.bmiHeader.biWidth = paintw;
							bi.bmiHeader.biHeight = painth;
							bi.bmiHeader.biPlanes = 1;
							bi.bmiHeader.biBitCount = 32;
							bi.bmiHeader.biCompression = BI_RGB;
							bi.bmiHeader.biSizeImage = paintw * h * 4;

							SetDIBitsToDevice(hdc, paintx1, painty1, paintw, painth, 0, 0, 0, painth, mFrameBuffer.data(), &bi, DIB_RGB_COLORS);
						} else {
							SetBkColor(hdc, VDSwizzleU32(mGammaTable[0]) >> 8);

							const RECT r { paintx1, painty1, paintx2, painty2 };

							ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &r, L"", 0, nullptr);
						}
					}
				}
			}
		}

		mCanvas.EndDirect(ps);
	}
}

void ATUIPrinterGraphicalOutputWindow::OnDpiChanged(int dpi) {
	if (dpi > 0 && dpi < 10000) {
		if (mViewDpi != dpi)
			UpdateZoom(nullptr);
	}
}

void ATUIPrinterGraphicalOutputWindow::UpdateViewOrigin() {
	mViewOriginX = mViewCenterX - (float)(mViewWidthPixels  / 2) * mViewMMPerPixel;
	mViewOriginY = mViewCenterY - (float)(mViewHeightPixels / 2) * mViewMMPerPixel;
}

void ATUIPrinterGraphicalOutputWindow::ScrollByPixels(int dx, int dy) {
	if (dx || dy) {
		mViewCenterX -= mViewMMPerPixel * dx;
		mViewCenterY -= mViewMMPerPixel * dy;
		UpdateViewOrigin();

		mCanvas.Scroll(dx, dy);
	}
}

void ATUIPrinterGraphicalOutputWindow::SetZoom(float clicks) {
	SetZoom(clicks, vdpoint32(mViewWidthPixels / 2, mViewHeightPixels / 2));
}

void ATUIPrinterGraphicalOutputWindow::SetZoom(float clicks, const vdpoint32& centerPt) {
	float newZoom = std::clamp(clicks, kZoomMin, kZoomMax);

	if (fabsf(mZoomClicks - newZoom) < 0.0001f)
		return;

	mZoomClicks = newZoom;

	UpdateZoom(&centerPt);
}

void ATUIPrinterGraphicalOutputWindow::UpdateZoom(const vdpoint32* centerPt) {
	float basePixelsPerMM = (float)mViewDpi / 25.4f;
	float newPixelsPerMM = basePixelsPerMM * powf(2.0f, (float)mZoomClicks / 5.0f);

	if (fabsf(mViewPixelsPerMM - newPixelsPerMM) > 1e-3f*std::max(mViewPixelsPerMM, newPixelsPerMM)) {
		const float pivotDX = centerPt ? centerPt->x - mViewWidthPixels / 2 : 0.0f;
		const float pivotDY = centerPt ? centerPt->y - mViewHeightPixels / 2 : 0.0f;

		float anchorX = mViewCenterX + mViewMMPerPixel * pivotDX;
		float anchorY = mViewCenterY + mViewMMPerPixel * pivotDY;

		mViewPixelsPerMM = newPixelsPerMM;
		mViewMMPerPixel = 1.0f / mViewPixelsPerMM;

		mViewCenterX = anchorX - mViewMMPerPixel * pivotDX;
		mViewCenterY = anchorY - mViewMMPerPixel * pivotDY;
		UpdateViewOrigin();

		ForceFullInvalidation();
	}
}

void ATUIPrinterGraphicalOutputWindow::ForceFullInvalidation() {
	InvalidateRect(mhwnd, nullptr, false);
	if (mpOutput) {
		bool all;
		vdrect32f r;
		mpOutput->ExtractInvalidationRect(all, r);
	}
}

void ATUIPrinterGraphicalOutputWindow::ProcessPendingInvalidation() {
	mbInvalidationPending = false;

	if (mpOutput) {
		bool invAll;
		vdrect32f invRect;

		if (mpOutput->ExtractInvalidationRect(invAll, invRect)) {
			if (invAll)
				InvalidateRect(mhwnd, nullptr, false);
			else
				InvalidatePaperArea(invRect.left, invRect.top, invRect.right, invRect.bottom);
		}
	}
}

void ATUIPrinterGraphicalOutputWindow::InvalidatePaperArea(float x1, float y1, float x2, float y2) {
	if (x1 >= x2 || y1 >= y2)
		return;

	x1 -= mViewOriginX;
	x2 -= mViewOriginX;
	y1 -= mViewOriginY;
	y2 -= mViewOriginY;

	x1 = x1 * mViewPixelsPerMM;
	y1 = y1 * mViewPixelsPerMM;
	x2 = x2 * mViewPixelsPerMM;
	y2 = y2 * mViewPixelsPerMM;

	if (x1 < 0)
		x1 = 0;

	if (y1 < 0)
		y1 = 0;

	if (x2 > (float)mViewWidthPixels)
		x2 = (float)mViewWidthPixels;

	if (y2 > (float)mViewHeightPixels)
		y2 = (float)mViewHeightPixels;

	if (x1 >= x2 || y1 >= y2)
		return;

	RECT r {
		(LONG)ceilf(x1 - 0.5f - 3.5f / 8.0f),
		(LONG)ceilf(y1 - 0.5f - 3.5f / 8.0f),
		(LONG)ceilf(x2 - 0.5f + 3.5f / 8.0f),
		(LONG)ceilf(y2 - 0.5f + 3.5f / 8.0f)
	};

	if (r.left < r.right && r.top < r.bottom)
		InvalidateRect(mhwnd, &r, FALSE);
}

vdrect32f ATUIPrinterGraphicalOutputWindow::GetDocumentBounds() const {
	if (mpOutput)
		return mpOutput->GetDocumentBounds();
	else
		return vdrect32f(0, 0, 0, 0);
}

bool ATUIPrinterGraphicalOutputWindow::Render(const ViewTransform& viewTransform, sint32 x, sint32 y, uint32 w, uint32 h, bool force) {
	// compute pixel bounding box
	const sint32 viewix1 = x;
	const sint32 viewiy1 = y;
	const sint32 viewix2 = x + w;
	const sint32 viewiy2 = y + h;
	const float viewfx1 = (float)viewix1;
	const float viewfy1 = (float)viewiy1;
	const float viewfx2 = (float)viewix2;
	const float viewfy2 = (float)viewiy2;

	// compute document space bounding box
	const float viewDocX1 = viewTransform.mOriginX + viewfx1 * viewTransform.mMMPerPixel;
	const float viewDocY1 = viewTransform.mOriginY + viewfy1 * viewTransform.mMMPerPixel;
	const float viewDocX2 = viewTransform.mOriginX + viewfx2 * viewTransform.mMMPerPixel;
	const float viewDocY2 = viewTransform.mOriginY + viewfy2 * viewTransform.mMMPerPixel;

	// pre-cull line range to entire vertical range
	ATPrinterGraphicalOutput::CullInfo cullInfo;

	if (mpOutput) {
		if (!mpOutput->PreCull(cullInfo, vdrect32f(viewDocX1, viewDocY1, viewDocX2, viewDocY2)) && !force)
			return false;
	}

	// clear framebuffer
	mFrameBuffer.resize(w * h);
	std::fill(mFrameBuffer.begin(), mFrameBuffer.end(), 0xFFFFFFFF);

	// initialize antialiasing buffer
	if (mABuffer.size() < w*8)
		mABuffer.resize(w*8);

	// render one scan line at a time
	const float viewSubPixelsPerMM = viewTransform.mPixelsPerMM * 8.0f;
	const float rowCenterToFirstSubRowOffset = viewTransform.mMMPerPixel * (-3.5f / 8.0f);
	const float subRowStep = viewTransform.mMMPerPixel / 8.0f;
	const float dotRadius = mDotSizeMM;
	const float dotRadiusSq = dotRadius * dotRadius;
	const sint32 subw = w * 8;

	for(uint32 yoff = 0; yoff < h; ++yoff) {
		const float docRowYC = viewDocY1 + ((float)yoff + 0.5f) * viewTransform.mMMPerPixel;
		const float docRowYD = 0.5f * viewTransform.mMMPerPixel;

		// pre-cull lines to scan line vertical range
		const float docRowY1 = docRowYC - docRowYD;
		const float docRowY2 = docRowYC + docRowYD;

		// pre-cull dots to scan line
		mDotCullBuffer.clear();
		if (mpOutput)
			mpOutput->ExtractNextLineDots(mDotCullBuffer, cullInfo, vdrect32f(viewDocX1, docRowY1, viewDocX2, docRowY2));

		// clear the antialiasing buffer
		uint8 *abuf = mABuffer.data();
		memset(abuf, 0, w * 8);

		// render all dots
		const float docRowYDPlusDotR = docRowYD + dotRadius;
		for(const RenderDot& dot : mDotCullBuffer) {
			float dy = dot.mY - docRowYC;

			// vertical cull to row
			if (fabsf(dy) >= docRowYDPlusDotR)
				continue;

			// offset vertical delta to first subrow
			dy += rowCenterToFirstSubRowOffset;

			// process one subrow at a time
			for(uint32 subRow = 0; subRow < 8; ++subRow, (dy += subRowStep)) {
				const float dx2 = dotRadiusSq - dy*dy;

				if (dx2 <= 0)
					continue;

				const float dx = sqrtf(dx2);

				// compute x range in document space
				const float xc = dot.mX - viewDocX1;
				const float x1 = xc - dx;
				const float x2 = xc + dx;

				// convert x range to subpixels
				const sint32 subx1 = (sint32)ceilf(x1 * viewSubPixelsPerMM - 0.5f);
				const sint32 subx2 = (sint32)ceilf(x2 * viewSubPixelsPerMM - 0.5f);

				// horizontally clip
				const sint32 csubx1 = std::max<sint32>(subx1, 0);
				const sint32 csubx2 = std::min<sint32>(subx2, subw);

				if (csubx1 >= csubx2)
					continue;

				// draw bits
				const uint32 ucsubx1 = (uint32)csubx1;
				const uint32 ucsubx2 = (uint32)csubx2;
				const uint32 maskx1 = ucsubx1 >> 3;
				const uint32 maskx2 = (ucsubx2 - 1) >> 3;
				const uint8 mask1 = 0xFF >> (ucsubx1 & 7);
				const uint8 mask2 = 0xFF << ((8 - ucsubx2) & 7);
				uint8 *__restrict dst = abuf + subRow + maskx1*8;

				if (maskx1 == maskx2) {
					*dst |= mask1 & mask2;
				} else {
					*dst |= mask1;
					dst += 8;

					for(uint32 x = maskx1 + 1; x < maskx2; ++x) {
						*dst = 0xFF;
						dst += 8;
					}

					*dst |= mask2;
				}
			}
		}

		// render antialiasing buffer row to framebuffer
		Downsample8x8(&mFrameBuffer[w * (h - 1 - yoff)], abuf, w);
	}

	return true;
}

void ATUIPrinterGraphicalOutputWindow::Downsample8x8(uint32 *dst, const uint8 *src, size_t w) {
#ifdef VD_CPU_X64
	if (VDCheckAllExtensionsEnabled(VDCPUF_SUPPORTS_POPCNT))
		return Downsample8x8_POPCNT64(dst, src, w);
#endif

#ifdef VD_CPU_X86
	if (VDCheckAllExtensionsEnabled(VDCPUF_SUPPORTS_POPCNT))
		return Downsample8x8_POPCNT32(dst, src, w);
#endif

	Downsample8x8_Scalar(dst, src, w);
}

void ATUIPrinterGraphicalOutputWindow::Downsample8x8_Scalar(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const uint32 level
			= mPopCnt8[asrc[0]]
			+ mPopCnt8[asrc[1]]
			+ mPopCnt8[asrc[2]]
			+ mPopCnt8[asrc[3]]
			+ mPopCnt8[asrc[4]]
			+ mPopCnt8[asrc[5]]
			+ mPopCnt8[asrc[6]]
			+ mPopCnt8[asrc[7]];
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}

#ifdef VD_CPU_X64
void ATUIPrinterGraphicalOutputWindow::Downsample8x8_POPCNT64(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const size_t level = _mm_popcnt_u64(VDReadUnalignedU64(asrc));
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}
#endif

#ifdef VD_CPU_X86
void ATUIPrinterGraphicalOutputWindow::Downsample8x8_POPCNT32(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const size_t level = _mm_popcnt_u32(VDReadUnalignedU32(asrc))
			+ _mm_popcnt_u32(VDReadUnalignedU32(asrc + 4));
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}
#endif

///////////////////////////////////////////////////////////////////////////////

ATPrinterOutputWindow::ATPrinterOutputWindow()
	: ATUIPaneWindow(kATUIPaneId_PrinterOutput, L"Printer Output")
	, mhwndTextEditor(NULL)
	, mLineBufIdx(0)
{
	mPreferredDockCode = kATContainerDockBottom;

	mAddedOutputFn = [this](ATPrinterOutput& output) { OnAddedOutput(output); };
	mRemovingOutputFn = [this](ATPrinterOutput& output) { OnRemovingOutput(output); };
	mAddedGraphicalOutputFn = [this](ATPrinterGraphicalOutput& output) { OnAddedGraphicalOutput(output); };
	mRemovingGraphicalOutputFn = [this](ATPrinterGraphicalOutput& output) { OnRemovingGraphicalOutput(output); };
}

ATPrinterOutputWindow::~ATPrinterOutputWindow() {
}

LRESULT ATPrinterOutputWindow::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_SIZE:
			OnSize();
			break;

		case WM_CONTEXTMENU:
			{
				int x = GET_X_LPARAM(lParam);
				int y = GET_Y_LPARAM(lParam);

				HMENU menu0 = LoadMenu(NULL, mpGraphicWindow ? MAKEINTRESOURCE(IDR_PRINTER_GRAPHIC_CONTEXT_MENU) : MAKEINTRESOURCE(IDR_PRINTER_CONTEXT_MENU));
				if (menu0) {
					HMENU menu = GetSubMenu(menu0, 0);
					BOOL cmd = 0;

					if (x == -1 && y == -1) {
						const vdpoint32& pt = mpGraphicWindow
							? mpGraphicWindow->TransformClientToScreen(vdpoint32(0, 0))
							: mpTextEditor->GetScreenPosForContextMenu();
						
						x = pt.x;
						y = pt.y;
					} else {
						POINT pt = {x, y};

						if (!mpGraphicWindow && ScreenToClient(mhwndTextEditor, &pt))
							mpTextEditor->SetCursorPixelPos(pt.x, pt.y);
					}

					cmd = TrackPopupMenu(menu, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_RETURNCMD, x, y, 0, mhwnd, NULL);

					DestroyMenu(menu0);

					switch(cmd) {
						case ID_CONTEXT_CLEAR:
							if (mpGraphicWindow)
								mpGraphicWindow->Clear();
							else if (mpTextEditor) {
								mpTextEditor->Clear();
								mLastTextOffset = 0;

								if (mpTextOutput) {
									mpTextOutput->Clear();
									mpTextOutput->Revalidate();
								}
							}
							break;

						case ID_CONTEXT_RESETVIEW:
							if (mpGraphicWindow)
								mpGraphicWindow->ResetView();
							break;

						case ID_SAVEAS_PNGIMAGE96:
							if (mpGraphicWindow)
								mpGraphicWindow->SaveAsPNG(96.0f);
							break;

						case ID_SAVEAS_PNGIMAGE300:
							if (mpGraphicWindow)
								mpGraphicWindow->SaveAsPNG(300.0f);
							break;
					}
				}
			}
			break;
	}

	return ATUIPaneWindow::WndProc(msg, wParam, lParam);
}

bool ATPrinterOutputWindow::OnCreate() {
	if (!ATUIPaneWindow::OnCreate())
		return false;

	if (!VDCreateTextEditor(~mpTextEditor))
		return false;

	mhwndTextEditor = (HWND)mpTextEditor->Create(WS_EX_NOPARENTNOTIFY, WS_CHILD|WS_VISIBLE, 0, 0, 0, 0, (VDGUIHandle)mhwnd, 100);

	OnFontsUpdated();

	mpTextEditor->SetReadOnly(true);

	OnSize();

	mpOutputMgr = static_cast<ATPrinterOutputManager *>(&g_sim.GetPrinterOutputManager());
	mpOutputMgr->OnAddedOutput.Add(&mAddedOutputFn);
	mpOutputMgr->OnRemovingOutput.Add(&mRemovingOutputFn);
	mpOutputMgr->OnAddedGraphicalOutput.Add(&mAddedGraphicalOutputFn);
	mpOutputMgr->OnRemovingGraphicalOutput.Add(&mRemovingGraphicalOutputFn);

	AttachToAnyOutput();

	return true;
}

void ATPrinterOutputWindow::OnDestroy() {
	DetachFromTextOutput();
	DetachFromGraphicsOutput();

	if (mpOutputMgr) {
		mpOutputMgr->OnAddedOutput.Remove(&mAddedOutputFn);
		mpOutputMgr->OnRemovingOutput.Remove(&mRemovingOutputFn);
		mpOutputMgr->OnAddedGraphicalOutput.Remove(&mAddedGraphicalOutputFn);
		mpOutputMgr->OnRemovingGraphicalOutput.Remove(&mRemovingGraphicalOutputFn);
		mpOutputMgr = nullptr;
	}

	ATUIPaneWindow::OnDestroy();
}

void ATPrinterOutputWindow::OnSize() {
	if (mpGraphicWindow) {
		mpGraphicWindow->SetArea(GetClientArea());
	} else if (mhwndTextEditor) {
		ATUINativeWindowProxy proxy(mhwndTextEditor);

		proxy.SetArea(GetClientArea());
	}
}

void ATPrinterOutputWindow::OnFontsUpdated() {
	if (mhwndTextEditor)
		SendMessage(mhwndTextEditor, WM_SETFONT, (WPARAM)ATGetConsoleFontW32(), TRUE);
}

void ATPrinterOutputWindow::OnSetFocus() {
	if (mpGraphicWindow)
		mpGraphicWindow->Focus();
	else
		::SetFocus(mhwndTextEditor);
}

void ATPrinterOutputWindow::OnAddedOutput(ATPrinterOutput& output) {
	if (!mpTextOutput && !mpGraphicsOutput)
		AttachToAnyOutput();
}

void ATPrinterOutputWindow::OnRemovingOutput(ATPrinterOutput& output) {
	if (&output == mpTextOutput) {
		DetachFromTextOutput();

		AttachToAnyOutput();
	}
}

void ATPrinterOutputWindow::OnAddedGraphicalOutput(ATPrinterGraphicalOutput& output) {
	if (!mpGraphicsOutput)
		AttachToAnyOutput();
}

void ATPrinterOutputWindow::OnRemovingGraphicalOutput(ATPrinterGraphicalOutput& output) {
	if (&output == mpGraphicsOutput) {
		DetachFromGraphicsOutput();

		AttachToAnyOutput();
	}
}

void ATPrinterOutputWindow::AttachToAnyOutput() {
	if (mpOutputMgr) {
		if (mpOutputMgr->GetGraphicalOutputCount() > 0)
			AttachToGraphicsOutput(mpOutputMgr->GetGraphicalOutput(0));
		else if (mpOutputMgr->GetOutputCount() > 0)
			AttachToTextOutput(mpOutputMgr->GetOutput(0));
	}
}

void ATPrinterOutputWindow::AttachToTextOutput(ATPrinterOutput& output) {
	DetachFromGraphicsOutput();

	mLastTextOffset = 0;
	if (mpTextEditor)
		mpTextEditor->Clear();

	mpTextOutput = &output;
	mpTextOutput->SetOnInvalidation(
		[this] {
			UpdateTextOutput();
		}
	);

	UpdateTextOutput();
}

void ATPrinterOutputWindow::DetachFromTextOutput() {
	if (mpTextOutput) {
		mpTextOutput->SetOnInvalidation(nullptr);
		mpTextOutput = nullptr;

		if (mpTextEditor)
			mpTextEditor->Clear();
	}
}

void ATPrinterOutputWindow::UpdateTextOutput() {
	if (mpTextOutput) {
		const size_t offset = mpTextOutput->GetLength();

		if (offset > mLastTextOffset) {
			if (mpTextEditor)
				mpTextEditor->Append(mpTextOutput->GetTextPointer(mLastTextOffset));

			mLastTextOffset = offset;
		}

		mpTextOutput->Revalidate();
	}
}

void ATPrinterOutputWindow::AttachToGraphicsOutput(ATPrinterGraphicalOutput& output) {
	DetachFromTextOutput();
	DetachFromGraphicsOutput();

	mpGraphicsOutput = &output;

	mpTextEditor->Clear();

	ATUINativeWindowProxy proxy(mhwndTextEditor);
	proxy.Hide();

	mpGraphicWindow = new ATUIPrinterGraphicalOutputWindow;
	mpGraphicWindow->CreateChild(mhdlg, 101, 0, 0, 0, 0, WS_CHILD | WS_VISIBLE | WS_TABSTOP);

	OnSize();

	mpGraphicWindow->AttachToOutput(output);
}

void ATPrinterOutputWindow::DetachFromGraphicsOutput() {
	if (mpGraphicsOutput) {
		mpGraphicsOutput->SetOnInvalidation(nullptr);
		mpGraphicsOutput = nullptr;
	}

	if (mpGraphicWindow) {
		mpGraphicWindow->Destroy();
		mpGraphicWindow = nullptr;

		// the text window was not resized while the graphic window existed, so fix that now
		OnSize();

		mpTextEditor->Clear();

		ATUINativeWindowProxy proxy(mhwndTextEditor);
		proxy.Show();
	}
}

////////////////////////////////////////////////////////////////////////////////

void ATUIDebuggerRegisterPrinterOutputPane() {
	ATRegisterUIPaneType(kATUIPaneId_PrinterOutput, VDRefCountObjectFactory<ATPrinterOutputWindow, ATUIPane>);
}
